/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group (THG).                                         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of aug_eos5. The full aug_eos5 copyright notice,      *
 * including terms governing use, modification, and redistribution, is       *
 * contained in the files COPYING and Copyright.html. COPYING can be         *
 * found at the root of the source code distribution tree; Copyright.html    *
 * can be found at the root level of an installed copy of the electronic     *
 * aug_eos5 document set and is linked from the top-level documents page.   *
 * It can also be found at                                                   *
 * http://www.hdfgroup.org/H4H5TOOLS/doc/Copyright.html. If you do not have  *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "hdf5.h"
#include "HE5_HdfEosDef.h"
#include "util.h"
#include "read.h"
#include "write.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define _C_VECTOR_IMPL
#include "cvct.h"


/** 
 * @brief one instance of this structure keeps all necessary information for each grid
 * collect_grid_aux() function fills this structure using the HDF-EOS5 library.
 * @see collect_grid_aux
 */
typedef struct {
  challoc gridname;
  grid_t grid;
  int orthogonal;
  int ydimmajor;
  doublealloc lons;
  doublealloc lats;
} grid_aug_t;

void init_grid_aug_t(grid_aug_t *e)
{
  challoc_init(&e->gridname);
  init_grid_t(&e->grid);
  doublealloc_init(&e->lons);
  doublealloc_init(&e->lats);
}
void free_grid_aug_t(grid_aug_t *e)
{
  challoc_free(&e->gridname);
  free_grid_t(&e->grid);
  doublealloc_free(&e->lons);
  doublealloc_free(&e->lats);
}
DECLARE_VECTOR_ALLOCATOR_FREE(gridaugalloc, grid_aug_t)

/** 
 * @brief one instance of this structure keeps all necessary information for each swath
 * collect_swath_aux() function fills this structure using the HDF-EOS5 library.
 * @see collect_swath_aux
 */
typedef struct {
  challoc swathname;
  swath_t swath;
} swath_aug_t;

void init_swath_aug_t(swath_aug_t *e)
{
  challoc_init(&e->swathname);
  init_swath_t(&e->swath);
}
void free_swath_aug_t(swath_aug_t *e)
{
  challoc_free(&e->swathname);
  free_swath_t(&e->swath);
}
DECLARE_VECTOR_ALLOCATOR_FREE(swathaugalloc, swath_aug_t)


static int shrink_lonlat(hid_t ctxid, const grid_t *grid, int ydimmajor, doublealloc *lons, doublealloc *lats)
{
  int succeeded = 0;

  do {
    int i;

    if (ydimmajor == 1) {
      for (i = 0; i < grid->info.ydim; ++i)
	lats->s[i] = lats->s[i * grid->info.xdim];
      lats->len = grid->info.ydim;
      lons->len = grid->info.xdim;
    }
    else if (ydimmajor == 0) {
      for (i = 0; i < grid->info.xdim; ++i)
	lons->s[i] = lons->s[i * grid->info.ydim];
      lons->len = grid->info.xdim;
      lats->len = grid->info.ydim;
    }
    else {
      ASSERT(0);
    }

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief retrieve all necessary information for the given grid.
 * 
 * @param ctxid context identifier
 * @param gridname grid name
 * @param[out] gridaux all necessary information for the given grid
 * 
 * @return 0 if successful
 */
static int collect_grid_aux(hid_t ctxid, const char *gridname, grid_aug_t *gridaux)
{
  doublealloc whole;
  int succeeded = 0;

  doublealloc_init(&whole);
  do {
    if (!challoc_copys(&gridaux->gridname, gridname)) nomem1_break("grid name")
    if (read_hdfeos5_onegrid(ctxid, gridname, &gridaux->grid) == -1) err1_break(3, "hdfeos5 read")

    /* After reading all information from the given file, it calculates
     * longitude and latitude values using an HDF-EOS5 function, and detects
     * the major dimension and whether or not one-dimensional coordinate
     * variable can be used instead of two-dimensional ones.
     */
    if (generate_lonlat(ctxid, &gridaux->grid, &gridaux->ydimmajor, &gridaux->lons, &gridaux->lats) == -1) err1_break(3, "generate lonlat")
    if ((gridaux->orthogonal = check_ortho_dimscale(&gridaux->grid, gridaux->ydimmajor, &gridaux->lons, &gridaux->lats)) == -1) err1_break(3, "check ortho")
    if (gridaux->orthogonal) {
      if (shrink_lonlat(ctxid, &gridaux->grid, gridaux->ydimmajor, &gridaux->lons, &gridaux->lats) == -1) err_break(3)
    }
    else {
      /* TODO: this has not been tested, yet. */
    }

    succeeded = 1;
  } while (0);

  if (1) {
    doublealloc_free(&whole);
  }
  return succeeded ? 0 : -1;
}

static int read_grid(hid_t ctxid, gridaugalloc *grids)
{
  int succeeded = 0;

  do {
    int i, failinside = 0;
    context_t *ctx = get_context(ctxid);

    for (i = 0; i < ctx->gridlist.len; ++i) {
      grid_aug_t *gridaux;

      if (!gridaugalloc_readyplus(grids, 1)) err1_flagbreak(3, failinside, "lonlat")
      gridaux = &grids->s[grids->len++];
      init_grid_aug_t(gridaux);

      if (collect_grid_aux(ctxid, ctx->gridlist.s[i].name.s, gridaux) == -1) err1_flagbreak(3, failinside, "lonlat")
      msg2("read grid", ctx->gridlist.s[i].name.s);
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief retrieve all necessary information for the given swath.
 * 
 * @param ctxid context identifier
 * @param swathname swath name
 * @param[out] swathaux all necessary information for the given swath
 * 
 * @return 0 if successful
 */
static int collect_swath_aux(hid_t ctxid, const char *swathname, swath_aug_t *swathaux)
{
  doublealloc wholelat;
  int succeeded = 0;

  doublealloc_init(&wholelat);
  do {
    if (!challoc_copys(&swathaux->swathname, swathname)) nomem1_break("swath name")
    if (read_hdfeos5_oneswath(ctxid, swathname, &swathaux->swath) == -1) err1_break(3, "hdfeos5 read")

    succeeded = 1;
  } while (0);

  if (1) {
    doublealloc_free(&wholelat);
  }
  return succeeded ? 0 : -1;
}

static int read_swath(hid_t ctxid, swathaugalloc *swaths)
{
  int succeeded = 0;

  do {
    int i, failinside = 0;
    context_t *ctx = get_context(ctxid);

    for (i = 0; i < ctx->swathlist.len; ++i) {
      swath_aug_t *swathaux;

      if (!swathaugalloc_readyplus(swaths, 1)) err1_flagbreak(3, failinside, "lonlat")
      swathaux = &swaths->s[swaths->len++];
      init_swath_aug_t(swathaux);

      if (collect_swath_aux(ctxid, ctx->swathlist.s[i].name.s, swathaux) == -1) err1_flagbreak(3, failinside, "lonlat")
      msg2("read swath", ctx->swathlist.s[i].name.s);

      if (swathaux->swath.dimmap.len > 0) err2_flagbreak(3, failinside, swathaux->swathname.s, "has a dimension map.")
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief augment one grid using the HDF5 library
 * 
 * @param ctxid context identifier
 * @param gridaux all necessary information retrieved using the HDF-EOS5 library
 * 
 * @return 0 if successful
 */
static int augment_onegrid(hid_t ctxid, const grid_aug_t *gridaux)
{
  challoc coordinates;
  int succeeded = 0;

  challoc_init(&coordinates);
  do {
    int i, failinside = 0;
    int xdimexplicit = 0, ydimexplicit = 0;
    context_t *ctx = get_context(ctxid);

    switch (ctx->augmentmethod) {
      case 1:
	break;
      case 2:
      case 3:
	if (create_aug_group(ctxid, gridaux->gridname.s, 1) == -1) err1_flagbreak(3, failinside, "create aug group")
	break;
      default:
	ASSERT(0);
    }
    if (failinside) break;

    /* Create an HDF5 dimension scale for each HDF-EOS5 dimension */
    for (i = 0; i < gridaux->grid.dim.len; ++i) {
      const dimension_t *dim = &gridaux->grid.dim.s[i];
      int isxdim, isydim, ispure = 1;

      isxdim = strcmp(dim->name.s, "XDim") == 0;
      isydim = strcmp(dim->name.s, "YDim") == 0;
      if ((isxdim || isydim) && gridaux->orthogonal)
	ispure = 0;

      if (write_hdf5_dimension(ctxid, gridaux->gridname.s, 1, ispure, dim) == -1) err1_flagbreak(3, failinside, "aug dimension")
      if (isxdim) xdimexplicit = 1;
      if (isydim) ydimexplicit = 1;
    }
    if (failinside) break;

    /* Create two HDF5 dimension scales if they are implicit in the HDF-EOS5 file */
    if (!xdimexplicit) {
      if (write_hdf5_implicit_dimension(ctxid, gridaux->gridname.s, &gridaux->grid, 1, !gridaux->orthogonal) == -1) err1_break(3, "aug x dimension")
    }
    if (!ydimexplicit) {
      if (write_hdf5_implicit_dimension(ctxid, gridaux->gridname.s, &gridaux->grid, 0, !gridaux->orthogonal) == -1) err1_break(3, "aug y dimension")
    }

    /* Associate created HDF5 dimension scales with corresponding HDF5 datasets */
    for (i = 0; i < gridaux->grid.field.len; ++i) {
      const fieldinfo_t *field = &gridaux->grid.field.s[i];
      if (associate_hdf5_dimension(ctxid, gridaux->gridname.s, 1, field, 1) == -1) err2_flagbreak(3, failinside, "assoc dimension", field->name.s)
    }
    if (failinside) break;

    if (gridaux->orthogonal) {
      /* Fill two special HDF5 dimension scales, XDim and YDim, with longitude and latitude values */
      if (fill_hdf5_1d_dimension(ctxid, gridaux->gridname.s, 1, &gridaux->lons) == -1) err1_break(3, "fill 1d lon")
      if (fill_hdf5_1d_dimension(ctxid, gridaux->gridname.s, 0, &gridaux->lats) == -1) err1_break(3, "fill 1d lat")

      /* Write attributes to conform to the CF Conventions */
      if (conform_cf_conventions_grid_1d(ctxid, gridaux->gridname.s, 1) == -1) err1_break(3, "conform to CF conventions lon")
      if (conform_cf_conventions_grid_1d(ctxid, gridaux->gridname.s, 0) == -1) err1_break(3, "conform to CF conventions lat")
    }
    else {
      /* Fill two coordinate variables with longitude and latitude values */
      if (fill_hdf5_2d_dimension(ctxid, gridaux->gridname.s, 1, &gridaux->lons, gridaux->ydimmajor, gridaux->grid.info.xdim, gridaux->grid.info.ydim) == -1) err1_break(3, "fill 2d lon")
      if (fill_hdf5_2d_dimension(ctxid, gridaux->gridname.s, 0, &gridaux->lats, gridaux->ydimmajor, gridaux->grid.info.xdim, gridaux->grid.info.ydim) == -1) err1_break(3, "fill 2d lat")

      /* Write attributes to conform to the CF Conventions */
      if (conform_cf_conventions_grid_2d_dim(ctxid, gridaux->gridname.s, 1) == -1) err1_break(3, "conform to CF conventions XDim")
      if (conform_cf_conventions_grid_2d_dim(ctxid, gridaux->gridname.s, 0) == -1) err1_break(3, "conform to CF conventions YDim")
      if (conform_cf_conventions_grid_2d_coordvar(ctxid, gridaux->gridname.s, 1, &coordinates) == -1) err1_break(3, "conform to CF conventions lon")
      if (conform_cf_conventions_grid_2d_coordvar(ctxid, gridaux->gridname.s, 0, &coordinates) == -1) err1_break(3, "conform to CF conventions lat")
      /* Write "coordinates" attribute for each data field */
      for (i = 0; i < gridaux->grid.field.len; ++i) {
	const fieldinfo_t *field = &gridaux->grid.field.s[i];
	if (conform_cf_conventions_2d_data(ctxid, gridaux->gridname.s, 1, field, &coordinates) == -1) err2_flagbreak(3, failinside, "conform to CF conventions", field->name.s)
      }
      if (failinside) break;
    }

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&coordinates);
  }
  return succeeded ? 0 : -1;
}

static int augment_grids(hid_t ctxid, const gridaugalloc *grids)
{
  int succeeded = 0;

  do {
    int i, failinside = 0;

    for (i = 0; i < grids->len; ++i) {
      const grid_aug_t *gridaux = &grids->s[i];
      msg_push(gridaux->gridname.s);
      if (augment_onegrid(ctxid, gridaux) == -1) err1_flagbreak(3, failinside, "aug onegrid")
      msg_pop();
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief augment one swath using the HDF5 library
 * 
 * @param ctxid context identifier
 * @param swathaux all necessary information retrieved using the HDF-EOS5 library
 * 
 * @return 0 if successful
 */
static int augment_oneswath(hid_t ctxid, const swath_aug_t *swathaux)
{
  challoc coordinates;
  int succeeded = 0;

  challoc_init(&coordinates);
  do {
    int i, failinside = 0;
    context_t *ctx = get_context(ctxid);

    switch (ctx->augmentmethod) {
      case 1:
	break;
      case 2:
      case 3:
	err1_flagbreak(3, failinside, "Swath is not supported, yet")
	break;
      default:
	ASSERT(0);
    }
    if (failinside) break;

    /* Create an HDF5 dimension scale for each HDF-EOS5 dimension */
    for (i = 0; i < swathaux->swath.dim.len; ++i) {
      const dimension_t *dim = &swathaux->swath.dim.s[i];
      if (write_hdf5_dimension(ctxid, swathaux->swathname.s, 0, 1, dim) == -1) err1_flagbreak(3, failinside, "aug dimension")
    }
    if (failinside) break;

    /* Associate created HDF5 dimension scales with corresponding HDF5 datasets originated from HDF-EOS5 data fields */
    for (i = 0; i < swathaux->swath.datafield.len; ++i) {
      const fieldinfo_t *datafield = &swathaux->swath.datafield.s[i];
      if (associate_hdf5_dimension(ctxid, swathaux->swathname.s, 0, datafield, 1) == -1) err1_flagbreak(3, failinside, "assoc dimension")
    }
    if (failinside) break;

    /* Associate created HDF5 dimension scales with corresponding HDF5 datasets originated from HDF-EOS5 geo-location fields */
    for (i = 0; i < swathaux->swath.geofield.len; ++i) {
      const fieldinfo_t *geofield = &swathaux->swath.geofield.s[i];
      if (associate_hdf5_dimension(ctxid, swathaux->swathname.s, 0, geofield, 0) == -1) err1_flagbreak(3, failinside, "assoc dimension")
    }
    if (failinside) break;

    /* At the time of writing this code, we couldn't find any good conventions
     * for HDF-EOS5 swaths. All swath files we have have one- or three- dimensional
     * Longitude and Latitude.
     */

    succeeded = 1;
  } while (0);

  if (1) {
    challoc_free(&coordinates);
  }
  return succeeded ? 0 : -1;
}

static int augment_swaths(hid_t ctxid, const swathaugalloc *swaths)
{
  int succeeded = 0;

  do {
    int i, failinside = 0;

    for (i = 0; i < swaths->len; ++i) {
      const swath_aug_t *swathaux = &swaths->s[i];
      msg_push(swathaux->swathname.s);
      if (augment_oneswath(ctxid, swathaux) == -1) err1_flagbreak(3, failinside, "aug oneswath")
      msg_pop();
    }
    if (failinside) break;

    succeeded = 1;
  } while (0);

  return succeeded ? 0 : -1;
}

/** 
 * @brief augment the given file with the given method.
 * The augmentation procedure has two steps:
 *   1. Retrieving all information using the HDF-EOS5 library
 *   2. Augmenting the given file using the HDF5 library with information retrieved in the previous step
 * 
 * @param fname filename to be augmented
 * @param augmentmethod 1, 2, or 3 is available
 * 
 * @return 0 if successful
 */
static int doit(const char *fname, int augmentmethod)
{
  hid_t ctxid = -1;
  gridaugalloc grids;
  swathaugalloc swaths;
  int succeeded = 0;

  gridaugalloc_init(&grids);
  swathaugalloc_init(&swaths);
  do {
    context_t *ctx = NULL;

    if ((ctxid = create_context(augmentmethod)) == -1) earlyerr1_break(3, "create context")
    if ((ctx = get_context(ctxid)) == NULL) earlyerr1_break(3, "internal error?")

    /* Step 1: collect all information using the HDF-EOS5 library */
    msg_push("read");
    if (open_hdfeos5(ctxid, fname) == -1) err1_break(3, "open hdfeos5 file")
    if (read_grid(ctxid, &grids) == -1) err1_break(3, "read hdfeos5 grid")
    if (read_swath(ctxid, &swaths) == -1) err1_break(3, "read hdfeos5 swath")
    if (close_hdfeos5(ctxid) == -1) err1_break(3, "close hdfeos5")
    msg_pop();

    /* Step 2: augment the given file using the HDF5 library */
    msg_push("augment");
    if (open_hdf5(ctxid, fname) == -1) err1_break(3, "open hdf5 file")
    if (augment_grids(ctxid, &grids) == -1) err1_break(3, "augment grid")
    if (augment_swaths(ctxid, &swaths) == -1) err1_break(3, "augment swath")
    if (close_hdf5(ctxid) == -1) err1_break(3, "close hdf5")
    msg_pop();

    succeeded = 1;
  } while (0);

  if (1) {
    gridaugalloc_free(&grids);
    swathaugalloc_free(&swaths);
    close_context(ctxid);
  }
  return succeeded ? 0 : -1;
}

static void usage()
{
  printf("Usage: aug_eos5 <file>\n");
  printf("  This program modifies <file>.\n");
}

int main(int argc, char **argv)
{
  int augmentmethod = 1;
  int c;

  while ((c = getopt(argc, argv, "m:")) != -1) {
    switch (c) {
      case 'm':
	augmentmethod = atoi(optarg);
	if (!(1 <= augmentmethod && augmentmethod <= 3)) return 1;
	break;
    }
  }

  if (optind + 1 != argc) {
    usage();
    return 1;
  }

  return doit(argv[optind], augmentmethod);
}

// vim:ts=8:sw=2:sts=2:cindent
